import numpy as np

from finesse import Model
import finesse.components as fc
import finesse.analysis.actions as fa


def test_changing_parameters():
    """Tests that masses and moments of inertia can change during a simulation and give
    the correct plants."""
    model = Model()
    model.add(fc.Mirror("ETM", T=0, L=0))
    model.add(fc.Mirror("ITM", T=0.014, L=0))
    model.add(fc.FreeMass("ETM_sus", model.ETM.mech))
    model.add(fc.Pendulum("ITM_sus", model.ITM.mech))

    F0_Hz = 2
    inertia = 10
    Q = 80
    F_Hz = 5
    dofs = ["z", "pitch", "yaw"]
    props = ["mass", "I_pitch", "I_yaw"]
    for dof in dofs:
        model.set(f"ITM_sus.f{dof}", F0_Hz)
        model.set(f"ITM_sus.Q{dof}", Q)

    to = [f"{tm}_sus.mech.{dof}" for tm in ["ETM", "ITM"] for dof in dofs]
    fr = [f"{tm}_sus.mech.F_{dof}" for tm in ["ETM", "ITM"] for dof in dofs]
    change = {f"{tm}_sus.{p}": inertia for tm in ["ETM", "ITM"] for p in props}

    sol = model.run(
        fa.Series(
            fa.FrequencyResponse([F_Hz], fr, to, name="fresp1"),
            fa.Change(change),
            fa.FrequencyResponse([F_Hz], fr, to, name="fresp2"),
        )
    )

    assert np.all(sol["fresp1"].out == 0)
    free_mass_sus = -1 / (inertia * (2 * np.pi * F_Hz) ** 2)
    pendulum_sus = (
        1 / (inertia * (2 * np.pi) ** 2) / (F0_Hz**2 - F_Hz**2 + 1j * F_Hz * F0_Hz / Q)
    )
    assert np.allclose(np.diag(sol["fresp2"].out[0, :3, :3]), free_mass_sus)
    assert np.allclose(np.diag(sol["fresp2"].out[0, 3:, 3:]), pendulum_sus)
